use std::mem::size_of;

use zerocopy::{AsBytes, FromBytes, Unaligned};

use crate::{Db, FilePointer, FileRange, Header, PAGE_SIZE, U32};

enum SlabKind {
    SplitFreeList,
    RelativeFreeList,
    AbsoluteFreeList,
}

impl SlabKind {
    fn for_size(size: u32) -> Self {
        if size == 1 {
            Self::SplitFreeList
        } else if size < 8 {
            Self::RelativeFreeList
        } else if (size as u64) <= PAGE_SIZE {
            Self::AbsoluteFreeList
        } else {
            panic!("invalid size")
        }
    }
}

// Best bitmap sizes in bytes for a 4096 Byte slab
// const BITMAP_SIZE: [u32; 8] = [456, 241, 164, 125, 100, 84, 72, 63];

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct FreeList {
    head: FilePointer,
}

impl FreeList {
    pub fn empty() -> Self {
        Self {
            head: FilePointer::null(),
        }
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct AllocatorState {
    pub general: FreeList,
    pub slabs: SlabListPointer,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct SlabListHeader {
    next: SlabListPointer,
    len: U32,
    size: U32,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct SlabListPointer(pub FilePointer);

impl SlabListHeader {
    pub fn capacity(&self) -> u32 {
        (self.size.get() - size_of::<SlabListHeader>() as u32) / size_of::<Slab>() as u32
    }
}

impl SlabListPointer {
    pub fn set_next(&self, db: &mut Db, next: SlabListPointer) {
        db.write(self.0, next);
    }

    pub fn set_len(&self, db: &mut Db, len: u32) {
        db.write(self.0 + size_of::<SlabListPointer>() as u64, U32::from(len));
    }

    pub fn init(&self, db: &mut Db, size: u32) {
        db.write(
            self.0,
            SlabListHeader {
                next: SlabListPointer(FilePointer::null()),
                size: size.into(),
                len: 0.into(),
            },
        );
    }

    pub fn ptr(&self, db: &Db, i: u32) -> FilePointer {
        let this: SlabListHeader = db.read(self.0);
        assert!(i < this.len.get());
        self.0 + size_of::<SlabListHeader>() as u64 + i as u64 * size_of::<Slab>() as u64
    }

    pub fn write(&self, db: &mut Db, i: u32, value: Slab) {
        let ptr = self.ptr(db, i);
        db.write(ptr, value);
    }

    pub fn get(&self, db: &Db, i: u32) -> SlabPointer {
        let ptr = self.ptr(db, i);
        SlabPointer(ptr)
    }

    pub fn add(&self, db: &mut Db, slab_size: u32) -> SlabPointer {
        let this: SlabListHeader = db.read(self.0);
        let capacity = this.capacity();
        let SlabListHeader { mut next, len, .. } = this;

        if len.get() >= capacity {
            if next.0 == FilePointer::null() {
                next = SlabListPointer(db.add_pages(1));
                next.init(db, PAGE_SIZE as u32);
                self.set_next(db, next);
            }

            return next.add(db, slab_size);
        }

        let len = len.get();
        self.set_len(db, len + 1);
        self.write(
            db,
            len,
            Slab {
                head: FilePointer::null(),
                size: slab_size.into(),
            },
        );

        SlabPointer(self.ptr(db, len))
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
pub struct Slab {
    head: FilePointer,
    size: U32,
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(transparent)]
pub struct SlabPointer(FilePointer);

impl SlabPointer {
    pub fn read(&self, db: &Db) -> Slab {
        db.read(self.0)
    }

    pub fn get(&self, db: &mut Db) -> FileRange {
        let Slab { head, size } = self.read(db);

        let size = size.get();

        match SlabKind::for_size(size) {
            SlabKind::SplitFreeList => todo!(),
            SlabKind::RelativeFreeList => todo!(),
            SlabKind::AbsoluteFreeList => {
                let mut next = head;

                if next == FilePointer::null() {
                    next = self.allocate_page(db);
                }

                let new_next = db.read(next);

                self.set_next(db, new_next);

                next.range(size as u64)
            }
        }
    }

    pub fn set_next(&self, db: &mut Db, next: FilePointer) {
        db.write(self.0, next);
    }

    pub fn allocate_page(&self, db: &mut Db) -> FilePointer {
        let Slab { head, size } = self.read(db);

        let size = size.get();

        match SlabKind::for_size(size) {
            SlabKind::SplitFreeList => todo!(),
            SlabKind::RelativeFreeList => todo!(),
            SlabKind::AbsoluteFreeList => {
                let n = PAGE_SIZE / size as u64;

                let page = db.add_pages(1);

                let mut next = head;
                for i in (0..n).rev() {
                    let current = page + i * size as u64;
                    db.write(current, next);
                    next = current;
                }

                self.set_next(db, next);

                next
            }
        }
    }
}
