use std::{borrow::BorrowMut, collections::HashMap, fs::File, mem::size_of, ops::Range, sync::Arc};

mod allocator;
mod atomic_arc;

use allocator::{AllocatorState, FreeList, SlabListPointer, SlabPointer};
use atomic_arc::AtomicArc;
use memmap::{Mmap, MmapMut};
use zerocopy::{AsBytes, FromBytes, LayoutVerified, Unaligned, LE};

const PAGE_SIZE: u64 = 4096;

type U64 = zerocopy::byteorder::U64<LE>;
type U32 = zerocopy::byteorder::U32<LE>;

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned, Debug, Hash, PartialEq, Eq)]
#[repr(transparent)]
pub struct FilePointer(U64);

impl FilePointer {
    fn page(n: u64) -> Self {
        Self((n * PAGE_SIZE).into())
    }
    fn null() -> Self {
        Self(U64::ZERO)
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned, Debug, PartialEq, Eq)]
#[repr(C)]
pub struct FileRange {
    start: FilePointer,
    len: U64,
}

impl std::ops::Add<u64> for FilePointer {
    type Output = Self;

    fn add(self, rhs: u64) -> Self::Output {
        Self((self.0.get() + rhs).into())
    }
}

impl FilePointer {
    pub fn range(&self, len: u64) -> FileRange {
        FileRange {
            start: *self,
            len: U64::new(len),
        }
    }
}

impl FileRange {
    fn len(&self) -> u64 {
        self.len.get()
    }
    fn as_range(&self) -> Range<usize> {
        let start = self.start.0.get().try_into().unwrap();
        start..start + usize::try_from(self.len()).unwrap()
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
struct Header {
    magic: [u8; 16],
    root: FilePointer,
    allocator_state: AllocatorState,
}

impl Default for Header {
    fn default() -> Self {
        Self {
            magic: *b"cool db format 1",
            root: FilePointer::null(),
            allocator_state: AllocatorState {
                general: FreeList::empty(),
                slabs: SlabListPointer(FilePointer::page(0) + size_of::<Header>() as u64),
            },
        }
    }
}

struct Snapshot {
    root: FilePointer,
    map: Mmap,
}

pub struct Reader {
    state: Arc<AtomicArc<Snapshot>>,
}

pub struct Db {
    file: File,
    map: MmapMut,
    header: Header,
    state: Arc<AtomicArc<Snapshot>>,
}

#[derive(Clone, Copy)]
struct Replaced {
    from: FileRange,
    to: Option<FileRange>,
}

pub struct TransactionHandle<'t> {
    db: &'t mut Db,
    replaced: HashMap<FilePointer, Replaced>,
    new: HashMap<FilePointer, FileRange>,
}

impl<'t> TransactionHandle<'t> {
    unsafe fn get_data(&mut self, range: FileRange) -> &mut [u8] {
        &mut self.db.map[range.as_range()]
    }

    pub unsafe fn modify(&mut self, range: FileRange) -> (FileRange, &mut [u8]) {
        let new = if let Some(&replaced) = self.replaced.get(&range.start) {
            assert_eq!(replaced.from, range);

            if let Some(to) = replaced.to {
                assert_eq!(to, range);
                to
            } else {
                let (to, _) = self.allocate(range.len());

                self.replaced.get_mut(&range.start).unwrap().to = Some(to);

                to
            }
        } else if let Some(&new) = self.new.get(&range.start) {
            assert_eq!(new, range);
            new
        } else {
            let (new, _) = self.allocate(range.len());

            let res = self.replaced.insert(
                new.start,
                Replaced {
                    from: range,
                    to: Some(new),
                },
            );
            debug_assert!(res.is_none());

            new
        };

        (new, self.get_data(new))
    }

    pub fn allocate(&mut self, length: u64) -> (FileRange, &mut [u8]) {
        unsafe {
            let new = self.allocate_range(length);

            let res = self.new.insert(new.start, new);
            debug_assert!(res.is_none());

            (new, self.get_data(new))
        }
    }

    pub fn free(&mut self, range: FileRange) {
        //
    }

    fn allocate_range(&mut self, length: u64) -> FileRange {
        let range: FileRange = todo!();

        range
    }
}

impl Db {
    pub fn add_slab(&mut self, size: u32) -> SlabPointer {
        let allocator_state = self.header.allocator_state;
        allocator_state.slabs.add(self, size)
    }

    fn transaction(f: impl FnOnce(TransactionHandle)) {
        // let handle = TransactionHandle {};
    }

    fn header_ptr() -> FilePointer {
        FilePointer(0.into())
    }

    fn root_ptr() -> FilePointer {
        FilePointer(16.into())
    }

    pub fn create_reader(&self) -> Reader {
        Reader {
            state: self.state.clone(),
        }
    }

    fn update_root(&mut self, new_root: FilePointer) -> Arc<Snapshot> {
        // TODO: we could write some here + flush here for better consistency
        // e.g. a copy of the new root pointer

        // flush all data in file
        self.map.flush().unwrap();

        // update root pointer and immediately flush
        self.write(Self::root_ptr(), new_root);
        self.map
            .flush_range(Self::root_ptr().0.get() as usize, size_of::<FilePointer>())
            .unwrap();

        // update data that readers see
        self.state.swap(Arc::new(Snapshot {
            root: new_root,
            map: self.create_readonly_map(),
        }))
    }

    fn read<T: FromBytes>(&self, at: FilePointer) -> T {
        self.read_range(at.range(size_of::<T>() as u64))
    }

    fn read_range<T: FromBytes>(&self, range: FileRange) -> T {
        LayoutVerified::<_, T>::new(&self.map[range.as_range()])
            .unwrap()
            .read()
    }

    fn write<T: AsBytes>(&mut self, at: FilePointer, data: T) {
        self.write_range(at.range(size_of::<T>() as u64), data)
    }

    fn write_range<T: AsBytes>(&mut self, range: FileRange, data: T) {
        LayoutVerified::<_, T>::new(&mut self.map[range.as_range()])
            .unwrap()
            .write(data)
    }

    fn modify<T: FromBytes + AsBytes>(&mut self, at: FilePointer) -> &mut T {
        self.modify_range(at.range(size_of::<T>() as u64))
    }

    fn modify_range<T: FromBytes + AsBytes>(&mut self, range: FileRange) -> &mut T {
        LayoutVerified::<_, T>::new(&mut self.map[range.as_range()])
            .unwrap()
            .into_mut()
    }

    fn reference<T: FromBytes + AsBytes>(&self, at: FilePointer) -> &T {
        self.reference_range(at.range(size_of::<T>() as u64))
    }

    fn reference_range<T: FromBytes + AsBytes>(&self, range: FileRange) -> &T {
        LayoutVerified::<_, T>::new(&self.map[range.as_range()])
            .unwrap()
            .into_ref()
    }

    fn remap(&mut self) {
        let map = unsafe { MmapMut::map_mut(&self.file) }.unwrap();
        self.map = map;
    }

    fn create_readonly_map(&self) -> Mmap {
        unsafe { Mmap::map(&self.file) }.unwrap()
    }

    fn add_pages(&mut self, n: u64) -> FilePointer {
        let len = self.file.metadata().unwrap().len();
        self.file.set_len(len + PAGE_SIZE * n).unwrap();

        self.remap();

        FilePointer::null() + len
    }

    pub fn new(file: File) -> Self {
        let len = file.metadata().unwrap().len();
        if len == 0 {
            file.set_len(PAGE_SIZE).unwrap();
        }

        let map = unsafe { MmapMut::map_mut(&file) }.unwrap();

        let mut db = Self {
            state: Arc::new(AtomicArc::new(Arc::new(Snapshot {
                root: FilePointer::null(),
                map: unsafe { Mmap::map(&file).unwrap() },
            }))),
            file,
            map,
            header: Header::default(),
        };

        if len == 0 {
            db.init_allocator();
            db.write(Self::header_ptr(), db.header);
        } else {
            db.header = db.read(Self::header_ptr());
        }

        let _ = db.state.swap(Arc::new(Snapshot {
            root: db.header.root,
            map: unsafe { Mmap::map(&db.file).unwrap() },
        }));

        db
    }

    fn init_allocator(&mut self) {
        let allocator_state = self.header.allocator_state;
        allocator_state.slabs.init(
            self,
            (PAGE_SIZE - size_of::<Header>() as u64).try_into().unwrap(),
        );
    }
}

#[cfg(test)]
mod tests {
    use std::io::Write;
    use std::process::Stdio;

    use super::*;

    #[test]
    fn it_works() {
        let mut db = Db::new(tempfile::tempfile().unwrap());

        let slab = db.add_slab(16);

        for i in 1..520 {
            let range = slab.get(&mut db);
            let start = range.start.0.get();
            dbg!(start);

            let data = [0; 16].map(|_| i as u8);

            db.write_range(range, data);
        }

        let mut child = std::process::Command::new("hexdump")
            .arg("-C")
            .stdin(Stdio::piped())
            .stdout(Stdio::inherit())
            .spawn()
            .unwrap();

        let mut stdin = child.stdin.take().expect("failed to get stdin");

        stdin.write_all(db.map.as_bytes()).unwrap();

        std::mem::drop(stdin);

        child.wait().unwrap();
    }
}
